from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status
from rest_framework.filters import SearchFilter
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet
from member.models import TFJMUser, Authorization, MotivationLetter, Solution, Synthesis
from tournament.models import Team, Tournament, Pool

from .serializers import UserSerializer, TeamSerializer, TournamentSerializer, AuthorizationSerializer, \
    MotivationLetterSerializer, SolutionSerializer, SynthesisSerializer, PoolSerializer


class UserViewSet(ModelViewSet):
    """
    Display list of users.
    """
    queryset = TFJMUser.objects.all()
    serializer_class = UserSerializer
    filter_backends = [DjangoFilterBackend, SearchFilter]
    filterset_fields = ['id', 'first_name', 'last_name', 'email', 'gender', 'student_class', 'role', 'year', 'team',
                        'team__trigram', 'is_superuser', 'is_staff', 'is_active', ]
    search_fields = ['$first_name', '$last_name', ]


class TeamViewSet(ModelViewSet):
    """
    Display list of teams.
    """
    queryset = Team.objects.all()
    serializer_class = TeamSerializer
    filter_backends = [DjangoFilterBackend, SearchFilter]
    filterset_fields = ['name', 'trigram', 'validation_status', 'selected_for_final', 'access_code', 'tournament',
                        'year', ]
    search_fields = ['$name', 'trigram', ]


class TournamentViewSet(ModelViewSet):
    """
    Display list of tournaments.
    """
    queryset = Tournament.objects.all()
    serializer_class = TournamentSerializer
    filter_backends = [DjangoFilterBackend, SearchFilter]
    filterset_fields = ['name', 'size', 'price', 'date_start', 'date_end', 'final', 'organizers', 'year', ]
    search_fields = ['$name', ]


class AuthorizationViewSet(ModelViewSet):
    """
    Display list of authorizations.
    """
    queryset = Authorization.objects.all()
    serializer_class = AuthorizationSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['user', 'type', ]


class MotivationLetterViewSet(ModelViewSet):
    """
    Display list of motivation letters.
    """
    queryset = MotivationLetter.objects.all()
    serializer_class = MotivationLetterSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['team', 'team__trigram', ]


class SolutionViewSet(ModelViewSet):
    """
    Display list of solutions.
    """
    queryset = Solution.objects.all()
    serializer_class = SolutionSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['team', 'team__trigram', 'problem', ]


class SynthesisViewSet(ModelViewSet):
    """
    Display list of syntheses.
    """
    queryset = Synthesis.objects.all()
    serializer_class = SynthesisSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['team', 'team__trigram', 'source', 'round', ]


class PoolViewSet(ModelViewSet):
    """
    Display list of pools.
    If the request is a POST request and the format is "A;X;x;Y;y;Z;z;..." where A = 1 or 1 = 2,
     X, Y, Z, ... are team trigrams, x, y, z, ... are numbers of problems, then this is interpreted as a
     creation a pool for the round A with the solutions of problems x, y, z, ... of the teams X, Y, Z, ... respectively.
    """
    queryset = Pool.objects.all()
    serializer_class = PoolSerializer
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['teams', 'teams__trigram', 'round', ]

    def create(self, request, *args, **kwargs):
        data = request.data
        try:
            spl = data.split(";")
            if len(spl) >= 7:
                round = int(spl[0])
                teams = []
                solutions = []
                for i in range((len(spl) - 1) // 2):
                    trigram = spl[1 + 2 * i]
                    pb = int(spl[2 + 2 * i])
                    team = Team.objects.get(trigram=trigram)
                    solution = Solution.objects.get(team=team, problem=pb, final=team.selected_for_final)
                    teams.append(team)
                    solutions.append(solution)
                pool = Pool.objects.create(round=round)
                pool.teams.set(teams)
                pool.solutions.set(solutions)
                pool.save()
                serializer = PoolSerializer(pool)
                headers = self.get_success_headers(serializer.data)
                return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)
        except BaseException:  # JSON data
            pass
        return super().create(request, *args, **kwargs)